/************************************************************
 *  Martin Toepfer     ***     PROJECT : Lingvisto          *
 *  Wissmannstr. 5     ***                                  *
 *  D - 22041 Hamburg  ***     File    : file.c             *
 *  Germany            ***     Date    : 04.04.1997         *
 ************************************************************/
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include "datastructure.h"
#include "edit.h"
#include "error.h"
#include "file.h"
#include "xawfileselect.h"
#include "xawfunction.h"

/************************************************************
 *          global variables (in datastructure.c)           *
 ************************************************************/
extern struct lvo_file head;
extern NODE current;

/************************************************************
 *               Xaw function prototypes                    *
 ************************************************************/
static void SaveChanges(int mode);
static void FileExists(void);

/**********************************************************************
 *                      fileio fuctions                               *
 **********************************************************************/
/************************************************************
 * new_file : delete current file from memory               *
 ************************************************************/
void new_file(void) {

  /*
   * changes in old file -> ask user if she wants to save them
   */

  if (get_state() == CHANGED)
    SaveChanges(NEW);

  /*
   * no changes -> remove old file from memory
   */

  else {
    delete_all();
    set_filename(NULL);
   }
}

/************************************************************
 * save_as_file : get filename, then save file              *
 ************************************************************/
void save_as_file(void) {

  /*
   * if wordlist is not empty then save
   */

  if (get_wordlist() != NULL) {
    
    /*
     * get filename from fileselectorbox
     */

    SelectFile(SAVE);
    if (strcmp(get_filename(), "unnamed file") != 0) {
      
      /*
       * check if file already exists
       */

      if (open(get_filename(), O_RDONLY) != -1)
	FileExists();
      else {
	 
	/*
	 * try to create and save file
	 */
	 
	if (save_file() != 0)
	  restore_filename();
      }
    }
  }
  else
    lvo_error(SAVE_EMPTY_LIST);
}

/************************************************************
 * save_file : write file if possible                       *
 ************************************************************/
char save_file(void) {

  /*
   * check if wordlist is not empty
   */

  if (get_wordlist() != NULL) {
    
    /*
     * check if we know the filename
     */

    if (strcmp(get_filename(), "unnamed file") == 0) {
      save_as_file();                /* read in filename */
    }
    
    /*
     * check if user didn't cancel fileselectorbox
     */

    if (strcmp(get_filename(), "unnamed file") != 0) {
      int fd = creat(get_filename(), S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
      if (fd != -1) {                /* able to open file */
	 
	/*
	 * write file header '#lvo-1.0#'
	 */

	char buf[BUFSIZ];
	register unsigned short i = 10;
	NODE help;
	strcpy(buf, "#lvo-1.0#\n");

	/*
	 * write each node from wordlist
	 */

	for(help = get_wordlist(); help != NULL; help = help->next) {
	  register unsigned char j = (unsigned char) 0;

	  /*
	   * copy native word into buffer
	   */

	  do {

	    /*
	     * write buffer into file if necessary
	     */

	    if (i >= BUFSIZ) {
	      if (write(fd, buf, BUFSIZ) != BUFSIZ) {
		lvo_error(FILE_WRITE);
		return((char) -1);
	      }
	      else
		i = 0;
	    }

	    /*
	     * copy character
	     */

	    if (help->pair.native[j] != '\0')
	      buf[i++] = help->pair.native[j++];
	    else
	      buf[i++] = '\n';
	  } while (buf[i-1] != '\n');

	  /*
	   * copy foreign word into buffer
	   */
	  
	  j = (unsigned char) 0;
	  do {
	    
	    /*
	     * write buffer into file if necessary
	     */

	    if (i >= BUFSIZ) {
	      if (write(fd, buf, BUFSIZ) != BUFSIZ) {
		lvo_error(FILE_WRITE);
		return((char) -1);
	      }
	      else
		i = 0;
	    }

	    /*
	     * copy character
	     */
	    
	    if (help->pair.foreign[j] != '\0')
	      buf[i++] = help->pair.foreign[j++];
	    else
	      buf[i++] = '\n';
	  } while (buf[i-1] != '\n');
	}
	
	/*
	 * write buffer into file
	 */
	
	if (write(fd, buf, i) != i) {
	  lvo_error(FILE_WRITE);
	  return((char) -1);
	}
	close(fd);
	set_state(UNCHANGED);
      }
      else {
	lvo_error(FILE_OPEN);
	return((char) -1);
      }
    }
  }
  else {
    lvo_error(SAVE_EMPTY_LIST);
    return((char) -1);
  }
  
  /*
   * success
   */

  return((char) 0);
}

/************************************************************
 * read_file : read pairs of words                          *
 *                                                          *
 * fd : filedescriptor                                      *
 ************************************************************/
char read_file(int fd) {
  char buf[BUFSIZ];                          /* buffered IO */
  int len;
   
  /*
   * read '#lvo-1.0#'
   */
   
  if (read(fd, buf, 10) == 10) {
    buf[9] = '\0';
    if (strcmp(buf, "#lvo-1.0#") == 0) {
	    
      /*
       * read wordpairs
       */
      
      char word1[MAXLINE];
      char word2[MAXLINE];
      char *tmp = NULL;
      register unsigned char j = (unsigned char) 0;
      delete_all();              /* delete current wordlist */
	    
      /*
       * fill buffer
       */
	
      while ((len = read(fd, buf, BUFSIZ)) > 0) {
	register unsigned short i;
	
	/*
	 * read single word from buffer
	 */
	       
	for (i = 0; i < len; i++) {
	  if (buf[i] != '\n') {
	    word1[j++] = buf[i];
	    if (j >= MAXLINE) {
	      return((char) -1);
	    }
	  }
	  else {
		     
	    /*
	     * insert word in datastructure
	     */
		     
	    word1[j] = '\0';
	    
	    /*
	     * we just read one word, so lets get another
	     */
	    
	    if (tmp == NULL) {
	      strcpy(word2, word1);
	      tmp = word2;
	    }
	    
	    /*
	     * we read two words, so lets insert them
	     */
	    
	    else {
	      char *w1 = (char *) malloc(strlen(word1) + 1);
	      char *w2 = (char *) malloc(strlen(word2) + 1);
	      if ((w1 != NULL) && (w2 != NULL)) {
		strcpy(w1, word1);
		strcpy(w2, word2);
		insert_pair(w2, w1);
		tmp = NULL;
	      }
	      else
		return((char) -1);
	    }
	    j = (unsigned char) 0;
	  }
	}
      }
      
      /*
       * success
       */
      
      current = get_wordlist();
      set_state(UNCHANGED);
      return((char) 0);
    }
    
    /*
     * file is no .lvo file
     */
    
    else
      return((char) -1);
  }
  else {
    return ((char) -1);
  }
}

/************************************************************
 * load_file : read file if possible                        *
 ************************************************************/
char load_file(void) {

  /*
   * changes in old file -> ask user if she wants to save them
   */

  if (get_state() == CHANGED)
    SaveChanges(LOAD);

  /*
   * no changes -> load file from disk
   */

  else {
    int fd;
    SelectFile(LOAD);
    if (strcmp(get_filename(), "unnamed file") != 0) {
      fd = open(get_filename(), O_RDONLY);
      
      /*
       * able to open file -> lets read it in
       */
      
      if (fd != -1) {
	if (read_file(fd) == -1) {
	  restore_filename();
	  lvo_error(NO_LVO_FILE);
	  return((char) -1);
	}
	close(fd);
      }
      
      /*
       * unable to open file
       */
      
      else {
	restore_filename();
	lvo_error(FILE_OPEN);
	return((char) -1);
      }
    }
  }
  
  /*
   * success
   */
  
  return((char) 0);
}

/************************************************************
 * quit_file : exit file                                    *
 ************************************************************/
void quit_file(void) {

  /*
   * changes in old file -> ask user if she wants to save them
   */

  if (get_state() == CHANGED)
    SaveChanges(QUIT);
  
  /*
   * no changes -> quit program
   */

  else
    Quit();
}

/**********************************************************************
 *                          Xaw - Interface                           *
 **********************************************************************/
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>

/************************************************************
 *              global variables (in xaw.c)                 *
 ************************************************************/
extern Boolean b;

/************************************************************
 * SaveChanges : widget that asks the user if he wants to   *
 *               save his changes                           *
 ************************************************************/
/**************************************************
 *        callback of 'Save' button               *
 **************************************************/
/* ARGSUSED */
static void Save(Widget w, XtPointer x, XtPointer y) {
  int mode = (int) x;
  Destroy(w, x, y);
  
  /*
   * save file
   */
  
  save_file();
  
  /*
   * execute selected menu item
   */
  
  if (mode == NEW)
    new_file();
  else if (mode == LOAD)
    load_file();
  else if (mode == QUIT)
    quit_file();
  DisplayNodeNative(current);
  DisplayNodeForeign(current);
}

/**************************************************
 * callback of 'Ignore' button                    *
 **************************************************/
/* ARGSUSED */
static void Ignore(Widget w, XtPointer x, XtPointer y) {
  int mode = (int) x;
  Destroy(w, x, y);
  set_state(UNCHANGED);
  
  /*
   * execute selected menu item
   */
  
  if (mode == NEW)
    new_file();
  else if (mode == LOAD)
    load_file();
  else if (mode == QUIT)
    quit_file();
  DisplayNodeNative(current);
  DisplayNodeForeign(current);
}

/**************************************************
 * begin of SaveChanges()                         *
 **************************************************/
static void SaveChanges(int mode) {
  Widget psh, form, mess, save, cancel, ignore;
  char message[] = "WARNING !\n\nThe changes applied"\
    " to this file are not yet saved\n"\
    "Pressing 'Ignore' will cause"\
    " DATA LOSS !!!!\n\n";
  psh = ComputePsh();
  
  /*
   * message formWidget
   */
  
  form = XtVaCreateWidget("", formWidgetClass, psh, NULL);
  
  /*
   * message labelWidget
   */
  
  mess = XtVaCreateManagedWidget(message, labelWidgetClass, form, NULL);
  XtVaSetValues(mess, XtNborderWidth, 0, NULL);
  
  /*
   * save commandWidget
   */
  
  save = XtVaCreateManagedWidget("Save", commandWidgetClass, form, NULL);
  XtAddCallback(save, XtNcallback, Save, (XtPointer) mode);
  XtVaSetValues(save, XtNfromVert, mess, NULL);
  
  /*
   * cancel commandWidget
   */
  
  cancel = XtVaCreateManagedWidget("Cancel", commandWidgetClass, form, NULL);
  XtAddCallback(cancel, XtNcallback, Destroy, NULL);
  XtVaSetValues(cancel,
		XtNfromVert, mess,
		XtNfromHoriz, save,
		NULL);

  
  /*
   * ignore commandWidget
   */
  
  ignore = XtVaCreateManagedWidget("Ignore", commandWidgetClass, form, NULL);
  XtAddCallback(ignore, XtNcallback, Ignore, (XtPointer) mode);
  XtVaSetValues(ignore,
		XtNfromVert, mess,
		XtNfromHoriz, cancel,
		XtNhorizDistance, 50,
		NULL);
  
  /*
   * now lets popup the shell
   */
  
  XtManageChild(form);
  XtPopup(psh, XtGrabNonexclusive);
  DisplayState(get_filename());
}

/************************************************************
 * FileExists : asks user if he wants to destroy existing   *
 *              file                                        *
 ************************************************************/
/*************************************************
 *          callback of 'Yes' button             *
 *************************************************/
/* ARGSUSED */
static void Yes(Widget w, XtPointer x, XtPointer y) {
  b = (Boolean) False;
  Destroy(w, x, y);
  if (save_file() != 0)
    restore_filename();
}

/*************************************************
 *           callback of 'No' button             *
 *************************************************/
/* ARGSUSED */
static void No(Widget w, XtPointer x, XtPointer y) {
  b = (Boolean) False;
  Destroy(w, x, y);
  restore_filename();
}

/*************************************************
 *            begin of FileExists()              *
 *************************************************/
static void FileExists(void) {
  Widget psh, form, mess, yes, no;
  char message[] = "Warning !\n\nDo you really want"\
    " to write over an existing file ??\n"\
    "Pressing 'Yes'"\
    " might cause DATA LOSS !!!!\n\n";
  b = (Boolean) True;
  psh = ComputePsh();
  
  /*
   * message formWidget
   */
  
  form = XtVaCreateManagedWidget("", formWidgetClass, psh, NULL);
  
  /*
   * message text labelWidget
   */
  
  mess = XtVaCreateManagedWidget(message, labelWidgetClass, form, NULL);
  XtVaSetValues(mess, XtNborderWidth, 0, NULL);
  
  /*
   * yes commandWidget
   */
  
  yes = XtVaCreateManagedWidget("Yes", commandWidgetClass, form, NULL);
  XtAddCallback(yes, XtNcallback, Yes, NULL);
  XtVaSetValues(yes, XtNfromVert, mess, NULL);
  
  /*
   * no command widget
   */
  
  no = XtVaCreateManagedWidget("No", commandWidgetClass, form, NULL);
  XtAddCallback(no, XtNcallback, No, NULL);
  XtVaSetValues(no,
		XtNfromVert, mess,
		XtNfromHoriz, yes,
		NULL);
  
  /*
   * now lets popup the shell
   */
  
  XtManageChild(form);
  XtPopup(psh, XtGrabNonexclusive);
  DisplayState(get_filename());
  
  /*
   * wait until button is pressed
   */
  
  while (b == (Boolean) True) {
    DoNextEvent();
  }
}
